@@ -79,6 +79,7 @@ module Basement.String
     , readFloatingExact
     , upper
     , lower
+    , caseFold
     , isPrefixOf
     , isSuffixOf
     , isInfixOf
@@ -99,6 +100,7 @@ import qualified Basement.UArray           as Vec
 import qualified Basement.UArray           as C
 import qualified Basement.UArray.Mutable   as MVec
 import           Basement.Block.Mutable (Block(..), MutableBlock(..))
+import qualified Basement.Block.Mutable    as MBLK
 import           Basement.Compat.Bifunctor
 import           Basement.Compat.Base
 import           Basement.Compat.Natural
@@ -109,21 +111,22 @@ import           Basement.Numerical.Additive
 import           Basement.Numerical.Subtractive
 import           Basement.Numerical.Multiplicative
 import           Basement.Numerical.Number
+import           Basement.Cast
 import           Basement.Monad
 import           Basement.PrimType
 import           Basement.FinalPtr
 import           Basement.IntegralConv
 import           Basement.Floating
 import           Basement.MutableBuilder
+import           Basement.String.CaseMapping (upperMapping, lowerMapping, foldMapping)
 import           Basement.UTF8.Table
 import           Basement.UTF8.Helper
 import           Basement.UTF8.Base
 import           Basement.UTF8.Types
-import           Basement.UArray.Base as C (onBackendPrim, onBackend, offset, ValidRange(..), offsetsValidRange)
-import qualified Basement.Alg.Native.UTF8 as PrimBA
-import qualified Basement.Alg.Foreign.UTF8 as PrimAddr
-import qualified Basement.Alg.Native.String as BackendBA
-import qualified Basement.Alg.Foreign.String as BackendAddr
+import           Basement.UArray.Base as C (onBackendPrim, onBackend, onBackendPure, offset, ValidRange(..), offsetsValidRange, MUArray(..), MUArrayBackend(..))
+import           Basement.Alg.Class (Indexable)
+import qualified Basement.Alg.UTF8 as UTF8
+import qualified Basement.Alg.String as Alg
 import           GHC.Prim
 import           GHC.ST
 import           GHC.Types
@@ -165,9 +168,9 @@ validate array ofsStart sz = C.unsafeDewrap goBa goAddr array
   where
     unTranslateOffset start = first (\e -> e `offsetSub` start)
     goBa ba start =
-        unTranslateOffset start $ BackendBA.validate (start+end) ba (start + ofsStart)
-    goAddr (Ptr addr) start =
-        pure $ unTranslateOffset start $ BackendAddr.validate (start+end) addr (ofsStart + start)
+        unTranslateOffset start $ Alg.validate (start+end) ba (start + ofsStart)
+    goAddr ptr@(Ptr !_) start =
+        pure $ unTranslateOffset start $ Alg.validate (start+end) ptr (ofsStart + start)
     end = ofsStart `offsetPlusE` sz
 
 -- | Similar to 'validate' but works on a 'MutableByteArray'
@@ -358,8 +361,8 @@ splitAt n s@(String ba)
 indexN :: CountOf Char -> String -> Offset Word8
 indexN !n (String ba) = Vec.unsafeDewrap goVec goAddr ba
   where
-    goVec :: ByteArray# -> Offset Word8 -> Offset Word8
-    goVec !ma !start = loop start 0
+    goVec :: Block Word8 -> Offset Word8 -> Offset Word8
+    goVec (Block !ma) !start = loop start 0
       where
         !len = start `offsetPlusE` Vec.length ba
         loop :: Offset Word8 -> Offset Char -> Offset Word8
@@ -467,15 +470,15 @@ break predicate s@(String ba) = runST $ Vec.unsafeIndexer ba go
 breakEnd :: (Char -> Bool) -> String -> (String, String)
 breakEnd predicate s@(String arr)
     | k == end  = (s, mempty)
-    | otherwise = splitIndex k s
+    | otherwise = splitIndex (k `offsetSub` start) s
   where
     k = C.onBackend goVec (\_ -> pure . goAddr) arr
     (C.ValidRange !start !end) = offsetsValidRange arr
-    goVec (Block ba) = let k = BackendBA.revFindIndexPredicate predicate ba start end
-                        in if k == end then end else PrimBA.nextSkip ba k
-    goAddr (Ptr addr) =
-        let k = BackendAddr.revFindIndexPredicate predicate addr start end
-         in if k == end then end else PrimAddr.nextSkip addr k
+    goVec ba@(Block !_) = let k = Alg.revFindIndexPredicate predicate ba start end
+                        in if k == end then end else UTF8.nextSkip ba k
+    goAddr ptr@(Ptr !_) =
+        let k = Alg.revFindIndexPredicate predicate ptr start end
+         in if k == end then end else UTF8.nextSkip ptr k
 {-# INLINE [2] breakEnd #-}
 
 #if MIN_VERSION_base(4,9,0)
@@ -604,14 +607,14 @@ length (String arr)
     | otherwise    = C.onBackend goVec (\_ -> pure . goAddr) arr
   where
     (C.ValidRange !start !end) = offsetsValidRange arr
-    goVec (Block ma) = PrimBA.length ma start end
-    goAddr (Ptr ptr) = PrimAddr.length ptr start end
+    goVec ma = UTF8.length ma start end
+    goAddr ptr = UTF8.length ptr start end
 
 -- | Replicate a character @c@ @n@ times to create a string of length @n@
 replicate :: CountOf Char -> Char -> String
 replicate (CountOf n) c = runST (new nbBytes >>= fill)
   where
-    nbBytes   = scale (integralCast n :: Word) sz
+    nbBytes   = scale (cast n :: Word) sz
     sz = charToBytes (fromEnum c)
     fill :: PrimMonad prim => MutableString (PrimState prim) -> prim String
     fill ms = loop (Offset 0)
@@ -785,8 +788,8 @@ sortBy sortF s = fromList $ Data.List.sortBy sortF $ toList s -- FIXME for tests
 filter :: (Char -> Bool) -> String -> String
 filter predicate (String arr) = runST $ do
     (finalSize, dst) <- newNative sz $ \(MutableBlock mba) ->
-        C.onBackendPrim (\(Block ba) -> BackendBA.copyFilter predicate sz mba ba start)
-                        (\fptr -> withFinalPtr fptr $ \(Ptr addr) -> BackendAddr.copyFilter predicate sz mba addr start)
+        C.onBackendPrim (\ba@(Block !_) -> Alg.copyFilter predicate sz mba ba start)
+                        (\fptr -> withFinalPtr fptr $ \ptr@(Ptr !_) -> Alg.copyFilter predicate sz mba ptr start)
                         arr
     freezeShrink finalSize dst
   where
@@ -798,8 +801,8 @@ reverse :: String -> String
 reverse (String arr) = runST $ do
     ((), dst) <- newNative (C.length arr) $ \(MutableBlock mba) ->
             C.onBackendPrim
-                (\(Block ba) -> PrimBA.reverse mba 0 ba start end)
-                (\fptr -> withFinalPtr fptr $ \(Ptr addr) -> PrimAddr.reverse mba 0 addr start end)
+                (\ba@(Block !_) -> UTF8.reverse mba 0 ba start end)
+                (\fptr -> withFinalPtr fptr $ \ptr@(Ptr !_) -> UTF8.reverse mba 0 ptr start end)
                 arr
     freeze dst
   where
@@ -1052,7 +1055,7 @@ builderBuild sizeChunksI sb
 builderBuild_ :: PrimMonad m => Int -> Builder String MutableString Word8 m () () -> m String
 builderBuild_ sizeChunksI sb = either (\() -> internalError "impossible output") id <$> builderBuild sizeChunksI sb
 
-stringDewrap :: (ByteArray# -> Offset Word8 -> a)
+stringDewrap :: (Block Word8 -> Offset Word8 -> a)
              -> (Ptr Word8 -> Offset Word8 -> ST s a)
              -> String
              -> a
@@ -1065,18 +1068,18 @@ stringDewrap withBa withPtr (String ba) = C.unsafeDewrap withBa withPtr ba
 readIntegral :: (HasNegation i, IntegralUpsize Word8 i, Additive i, Multiplicative i, IsIntegral i) => String -> Maybe i
 readIntegral str
     | sz == 0   = Nothing
-    | otherwise = stringDewrap withBa (\(Ptr ptr) -> pure . withPtr ptr) str
+    | otherwise = stringDewrap withBa (\ptr@(Ptr !_) -> pure . withPtr ptr) str
   where
     !sz = size str
     withBa ba ofs =
-        let negativeSign = PrimBA.expectAscii ba ofs 0x2d
+        let negativeSign = UTF8.expectAscii ba ofs 0x2d
             startOfs     = if negativeSign then succ ofs else ofs
          in case decimalDigitsBA 0 ba endOfs startOfs of
                 (# acc, True, endOfs' #) | endOfs' > startOfs -> Just $! if negativeSign then negate acc else acc
                 _                                             -> Nothing
       where !endOfs = ofs `offsetPlusE` sz
     withPtr addr ofs =
-        let negativeSign = PrimAddr.expectAscii addr ofs 0x2d
+        let negativeSign = UTF8.expectAscii addr ofs 0x2d
             startOfs     = if negativeSign then succ ofs else ofs
          in case decimalDigitsPtr 0 addr endOfs startOfs of
                 (# acc, True, endOfs' #) | endOfs' > startOfs -> Just $! if negativeSign then negate acc else acc
@@ -1094,7 +1097,7 @@ readInteger = readIntegral
 readNatural :: String -> Maybe Natural
 readNatural str
     | sz == 0   = Nothing
-    | otherwise = stringDewrap withBa (\(Ptr ptr) -> pure . withPtr ptr) str
+    | otherwise = stringDewrap withBa (\ptr@(Ptr !_) -> pure . withPtr ptr) str
   where
     !sz = size str
     withBa ba stringStart =
@@ -1135,7 +1138,7 @@ readRational s =
         case mExponant of
             Just exponent
                 | exponent < -10000 || exponent > 10000 -> Nothing
-                | otherwise                             -> Just $ modF isNegative integral % (10 Prelude.^ (integralCast floatingDigits - exponent))
+                | otherwise                             -> Just $ modF isNegative integral % (10 Prelude.^ (cast floatingDigits - exponent))
             Nothing                                     -> Just $ modF isNegative integral % (10 Prelude.^ floatingDigits)
   where
     modF True  = negate . integralUpsize
@@ -1176,7 +1179,7 @@ readFloatingExact str f
     !sz = size str
 
     withBa ba stringStart =
-        let !isNegative = PrimBA.expectAscii ba stringStart 0x2d
+        let !isNegative = UTF8.expectAscii ba stringStart 0x2d
          in consumeIntegral isNegative (if isNegative then stringStart+1 else stringStart)
       where
         eofs = stringStart `offsetPlusE` sz
@@ -1184,7 +1187,7 @@ readFloatingExact str f
             case decimalDigitsBA 0 ba eofs startOfs of
                 (# acc, True , endOfs #) | endOfs > startOfs -> f isNegative acc 0 Nothing -- end of stream and no '.'
                 (# acc, False, endOfs #) | endOfs > startOfs ->
-                    if PrimBA.expectAscii ba endOfs 0x2e
+                    if UTF8.expectAscii ba endOfs 0x2e
                         then consumeFloat isNegative acc (endOfs + 1)
                         else consumeExponant isNegative acc 0 endOfs
                 _                                            -> Nothing
@@ -1192,31 +1195,31 @@ readFloatingExact str f
         consumeFloat isNegative integral startOfs =
             case decimalDigitsBA integral ba eofs startOfs of
                 (# acc, True, endOfs #) | endOfs > startOfs -> let (CountOf !diff) = endOfs - startOfs
-                                                                in f isNegative acc (integralCast diff) Nothing
+                                                                in f isNegative acc (cast diff) Nothing
                 (# acc, False, endOfs #) | endOfs > startOfs -> let (CountOf !diff) = endOfs - startOfs
-                                                                in consumeExponant isNegative acc (integralCast diff) endOfs
+                                                                in consumeExponant isNegative acc (cast diff) endOfs
                 _                                           -> Nothing
 
         consumeExponant !isNegative !integral !floatingDigits !startOfs
             | startOfs == eofs = f isNegative integral floatingDigits Nothing
             | otherwise        =
                 -- consume 'E' or 'e'
-                case PrimBA.nextAscii ba startOfs of
+                case UTF8.nextAscii ba startOfs of
                     StepASCII 0x45 -> consumeExponantSign (startOfs+1)
                     StepASCII 0x65 -> consumeExponantSign (startOfs+1)
                     _              -> Nothing
           where
             consumeExponantSign ofs
                 | ofs == eofs = Nothing
-                | otherwise   = let exponentNegative = PrimBA.expectAscii ba ofs 0x2d
+                | otherwise   = let exponentNegative = UTF8.expectAscii ba ofs 0x2d
                                  in consumeExponantNumber exponentNegative (if exponentNegative then ofs + 1 else ofs)
 
             consumeExponantNumber exponentNegative ofs =
                 case decimalDigitsBA 0 ba eofs ofs of
                     (# acc, True, endOfs #) | endOfs > ofs -> f isNegative integral floatingDigits (Just $! if exponentNegative then negate acc else acc)
                     _                                      -> Nothing
-    withPtr (Ptr ptr) stringStart = pure $
-        let !isNegative = PrimAddr.expectAscii ptr stringStart 0x2d
+    withPtr ptr@(Ptr !_) stringStart = pure $
+        let !isNegative = UTF8.expectAscii ptr stringStart 0x2d
          in consumeIntegral isNegative (if isNegative then stringStart+1 else stringStart)
       where
         eofs = stringStart `offsetPlusE` sz
@@ -1224,7 +1227,7 @@ readFloatingExact str f
             case decimalDigitsPtr 0 ptr eofs startOfs of
                 (# acc, True , endOfs #) | endOfs > startOfs -> f isNegative acc 0 Nothing -- end of stream and no '.'
                 (# acc, False, endOfs #) | endOfs > startOfs ->
-                    if PrimAddr.expectAscii ptr endOfs 0x2e
+                    if UTF8.expectAscii ptr endOfs 0x2e
                         then consumeFloat isNegative acc (endOfs + 1)
                         else consumeExponant isNegative acc 0 endOfs
                 _                                            -> Nothing
@@ -1232,23 +1235,23 @@ readFloatingExact str f
         consumeFloat isNegative integral startOfs =
             case decimalDigitsPtr integral ptr eofs startOfs of
                 (# acc, True, endOfs #) | endOfs > startOfs -> let (CountOf !diff) = endOfs - startOfs
-                                                                in f isNegative acc (integralCast diff) Nothing
+                                                                in f isNegative acc (cast diff) Nothing
                 (# acc, False, endOfs #) | endOfs > startOfs -> let (CountOf !diff) = endOfs - startOfs
-                                                                in consumeExponant isNegative acc (integralCast diff) endOfs
+                                                                in consumeExponant isNegative acc (cast diff) endOfs
                 _                                           -> Nothing
 
         consumeExponant !isNegative !integral !floatingDigits !startOfs
             | startOfs == eofs = f isNegative integral floatingDigits Nothing
             | otherwise        =
                 -- consume 'E' or 'e'
-                case PrimAddr.nextAscii ptr startOfs of
+                case UTF8.nextAscii ptr startOfs of
                     StepASCII 0x45 -> consumeExponantSign (startOfs+1)
                     StepASCII 0x65 -> consumeExponantSign (startOfs+1)
                     _              -> Nothing
           where
             consumeExponantSign ofs
                 | ofs == eofs = Nothing
-                | otherwise   = let exponentNegative = PrimAddr.expectAscii ptr ofs 0x2d
+                | otherwise   = let exponentNegative = UTF8.expectAscii ptr ofs 0x2d
                                  in consumeExponantNumber exponentNegative (if exponentNegative then ofs + 1 else ofs)
 
             consumeExponantNumber exponentNegative ofs =
@@ -1280,7 +1283,7 @@ readFloatingExact str f
 -- this function
 decimalDigitsBA :: (IntegralUpsize Word8 acc, Additive acc, Multiplicative acc, Integral acc)
                 => acc
-                -> ByteArray#
+                -> Block Word8
                 -> Offset Word8 -- end offset
                 -> Offset Word8 -- start offset
                 -> (# acc, Bool, Offset Word8 #)
@@ -1289,18 +1292,18 @@ decimalDigitsBA startAcc ba !endOfs !startOfs = loop startAcc startOfs
     loop !acc !ofs
         | ofs == endOfs = (# acc, True, ofs #)
         | otherwise     =
-            case PrimBA.nextAsciiDigit ba ofs of
+            case UTF8.nextAsciiDigit ba ofs of
                 sg@(StepDigit d) | isValidStepDigit sg -> loop (10 * acc + integralUpsize d) (succ ofs)
                                  | otherwise           -> (# acc, False, ofs #)
-{-# SPECIALIZE decimalDigitsBA :: Integer -> ByteArray# -> Offset Word8 -> Offset Word8 -> (# Integer, Bool, Offset Word8 #) #-}
-{-# SPECIALIZE decimalDigitsBA :: Natural -> ByteArray# -> Offset Word8 -> Offset Word8 -> (# Natural, Bool, Offset Word8 #) #-}
-{-# SPECIALIZE decimalDigitsBA :: Int -> ByteArray# -> Offset Word8 -> Offset Word8 -> (# Int, Bool, Offset Word8 #) #-}
-{-# SPECIALIZE decimalDigitsBA :: Word -> ByteArray# -> Offset Word8 -> Offset Word8 -> (# Word, Bool, Offset Word8 #) #-}
+{-# SPECIALIZE decimalDigitsBA :: Integer -> Block Word8 -> Offset Word8 -> Offset Word8 -> (# Integer, Bool, Offset Word8 #) #-}
+{-# SPECIALIZE decimalDigitsBA :: Natural -> Block Word8 -> Offset Word8 -> Offset Word8 -> (# Natural, Bool, Offset Word8 #) #-}
+{-# SPECIALIZE decimalDigitsBA :: Int -> Block Word8 -> Offset Word8 -> Offset Word8 -> (# Int, Bool, Offset Word8 #) #-}
+{-# SPECIALIZE decimalDigitsBA :: Word -> Block Word8 -> Offset Word8 -> Offset Word8 -> (# Word, Bool, Offset Word8 #) #-}
 
 -- | same as decimalDigitsBA specialized for ptr #
 decimalDigitsPtr :: (IntegralUpsize Word8 acc, Additive acc, Multiplicative acc, Integral acc)
                  => acc
-                 -> Addr#
+                 -> Ptr Word8
                  -> Offset Word8 -- end offset
                  -> Offset Word8 -- start offset
                  -> (# acc, Bool, Offset Word8 #)
@@ -1309,23 +1312,70 @@ decimalDigitsPtr startAcc ptr !endOfs !startOfs = loop startAcc startOfs
     loop !acc !ofs
         | ofs == endOfs = (# acc, True, ofs #)
         | otherwise     =
-            case PrimAddr.nextAsciiDigit ptr ofs of
+            case UTF8.nextAsciiDigit ptr ofs of
                 sg@(StepDigit d) | isValidStepDigit sg -> loop (10 * acc + integralUpsize d) (succ ofs)
                                  | otherwise           -> (# acc, False, ofs #)
-{-# SPECIALIZE decimalDigitsPtr :: Integer -> Addr# -> Offset Word8 -> Offset Word8 -> (# Integer, Bool, Offset Word8 #) #-}
-{-# SPECIALIZE decimalDigitsPtr :: Natural -> Addr# -> Offset Word8 -> Offset Word8 -> (# Natural, Bool, Offset Word8 #) #-}
-{-# SPECIALIZE decimalDigitsPtr :: Int -> Addr# -> Offset Word8 -> Offset Word8 -> (# Int, Bool, Offset Word8 #) #-}
-{-# SPECIALIZE decimalDigitsPtr :: Word -> Addr# -> Offset Word8 -> Offset Word8 -> (# Word, Bool, Offset Word8 #) #-}
+{-# SPECIALIZE decimalDigitsPtr :: Integer -> Ptr Word8 -> Offset Word8 -> Offset Word8 -> (# Integer, Bool, Offset Word8 #) #-}
+{-# SPECIALIZE decimalDigitsPtr :: Natural -> Ptr Word8 -> Offset Word8 -> Offset Word8 -> (# Natural, Bool, Offset Word8 #) #-}
+{-# SPECIALIZE decimalDigitsPtr :: Int -> Ptr Word8 -> Offset Word8 -> Offset Word8 -> (# Int, Bool, Offset Word8 #) #-}
+{-# SPECIALIZE decimalDigitsPtr :: Word -> Ptr Word8 -> Offset Word8 -> Offset Word8 -> (# Word, Bool, Offset Word8 #) #-}
+
+-- | Convert a 'String' 'Char' by 'Char' using a case mapping function.
+caseConvert :: (Char -> CM) -> String -> String
+caseConvert op s@(String arr) = runST $ do
+  mba <- MBLK.new iLen
+  nL <- C.onBackendPrim
+        (\blk  -> go mba blk (Offset 0) start)
+        (\fptr -> withFinalPtr fptr $ \ptr -> go mba ptr (Offset 0) start)
+        arr
+  freeze . MutableString $ MVec.MUArray 0 nL (C.MUArrayMBA mba)
+  where
+    !(C.ValidRange start end) = C.offsetsValidRange arr
+    !iLen = 1 + C.length arr
+    go :: (Indexable container Word8, PrimMonad prim)
+       => MutableBlock Word8 (PrimState prim)
+       -> container
+       -> Offset Word8
+       -> Offset Word8
+       -> prim (CountOf Word8)
+    go !dst !src = loop dst iLen 0
+      where
+        eSize !e = if e == '\0' then 0 else charToBytes (fromEnum e)
+        loop !dst !allocLen !nLen !dstIdx !srcIdx
+          | srcIdx == end = return nLen
+          | nLen == allocLen = realloc
+          | otherwise = do
+              let !(CM c1 c2 c3) = op c
+                  !(Step c nextSrcIdx) = UTF8.next src srcIdx
+              nextDstIdx <- UTF8.writeUTF8 dst dstIdx c1
+              if c2 == '\0' -- We keep the most common case loop as short as possible.
+                then loop dst allocLen (nLen + charToBytes (fromEnum c1)) nextDstIdx nextSrcIdx
+                else do
+                  let !cSize = eSize c1 + eSize c2 + eSize c3
+                  nextDstIdx <- UTF8.writeUTF8 dst nextDstIdx c2
+                  nextDstIdx <- if c3 == '\0' then return nextDstIdx else UTF8.writeUTF8 dst nextDstIdx c3
+                  loop dst allocLen (nLen + cSize) nextDstIdx nextSrcIdx
+          where
+            {-# NOINLINE realloc #-}
+            realloc = do
+              let nAll = allocLen + allocLen + 1
+              nDst <- MBLK.new nAll
+              MBLK.unsafeCopyElements nDst 0 dst 0 nLen
+              loop nDst nAll nLen dstIdx srcIdx
 
 -- | Convert a 'String' to the upper-case equivalent.
---   Does not properly support multicharacter Unicode conversions.
 upper :: String -> String
-upper = charMap toUpper
+upper = caseConvert upperMapping
 
 -- | Convert a 'String' to the upper-case equivalent.
---   Does not properly support multicharacter Unicode conversions.
 lower :: String -> String
-lower = charMap toLower
+lower = caseConvert lowerMapping
+
+-- | Convert a 'String' to the unicode case fold equivalent.
+--
+-- Case folding is mostly used for caseless comparison of strings.
+caseFold :: String -> String
+caseFold = caseConvert foldMapping
 
 -- | Check whether the first string is a prefix of the second string.
 isPrefixOf :: String -> String -> Bool
@@ -1374,15 +1424,15 @@ all :: (Char -> Bool) -> String -> Bool
 all predicate (String arr) = C.onBackend goBA (\_ -> pure . goAddr) arr
   where
     !(C.ValidRange start end) = C.offsetsValidRange arr
-    goBA (Block ba)   = PrimBA.all predicate ba start end
-    goAddr (Ptr addr) = PrimAddr.all predicate addr start end
+    goBA ba   = UTF8.all predicate ba start end
+    goAddr addr = UTF8.all predicate addr start end
 
 any :: (Char -> Bool) -> String -> Bool
 any predicate (String arr) = C.onBackend goBA (\_ -> pure . goAddr) arr
   where
     !(C.ValidRange start end) = C.offsetsValidRange arr
-    goBA (Block ba)   = PrimBA.any predicate ba start end
-    goAddr (Ptr addr) = PrimAddr.any predicate addr start end
+    goBA ba   = UTF8.any predicate ba start end
+    goAddr addr = UTF8.any predicate addr start end
 
 -- | Transform string @src@ to base64 binary representation.
 toBase64 :: String -> String
