@@ -106,7 +106,6 @@ module Basement.UArray
     , toBase64Internal
     ) where
 
-import           Control.Monad (when)
 import           GHC.Prim
 import           GHC.Types
 import           GHC.Word
@@ -136,8 +135,6 @@ import           Basement.MutableBuilder
 import           Basement.Bindings.Memory (sysHsMemFindByteBa, sysHsMemFindByteAddr)
 import qualified Basement.Compat.ExtList as List
 import qualified Basement.Base16 as Base16
-import qualified Basement.Alg.Native.Prim as PrimBA
-import qualified Basement.Alg.Foreign.Prim as PrimAddr
 import qualified Basement.Alg.Mutable as Alg
 import qualified Basement.Alg.Class as Alg
 import qualified Basement.Alg.PrimArray as Alg
@@ -159,6 +156,12 @@ foreignMem :: PrimType ty
            -> UArray ty
 foreignMem fptr nb = UArray (Offset 0) nb (UArrayAddr fptr)
 
+-- | Create a foreign UArray from foreign memory and given offset/size
+--
+-- No check are performed to make sure this is valid, so this is unsafe.
+--
+-- This is particularly useful when dealing with foreign memory and
+-- 'ByteString'
 fromForeignPtr :: PrimType ty
                => (ForeignPtr ty, Int, Int) -- ForeignPtr, an offset in prim elements, a size in prim elements
                -> UArray ty
@@ -188,6 +191,10 @@ unsafeCopyFrom v' newLen f = new newLen >>= fill 0 >>= unsafeFreeze
             | otherwise  = do f v' i r'
                               fill (i + 1) r'
 
+-- | Freeze a MUArray into a UArray by copying all the content is a pristine new buffer
+--
+-- The MUArray in parameter can be still be used after the call without
+-- changing the resulting frozen data.
 freeze :: (PrimType ty, PrimMonad prim) => MUArray ty (PrimState prim) -> prim (UArray ty)
 freeze ma = do
     ma' <- new len
@@ -195,8 +202,13 @@ freeze ma = do
     unsafeFreeze ma'
   where len = mutableLength ma
 
+-- | Just like 'freeze' but copy only the first n bytes
+--
+-- The size requested need to be smaller or equal to the length
+-- of the MUArray, otherwise a Out of Bounds exception is raised
 freezeShrink :: (PrimType ty, PrimMonad prim) => MUArray ty (PrimState prim) -> CountOf ty -> prim (UArray ty)
 freezeShrink ma n = do
+    when (n > mutableLength ma) $ primOutOfBound OOB_MemCopy (sizeAsOffset n) (mutableLength ma)
     ma' <- new n
     copyAt ma' (Offset 0) ma (Offset 0) n
     unsafeFreeze ma'
@@ -292,21 +304,22 @@ copyToPtr arr dst@(Ptr dst#) = onBackendPrim copyBa copyPtr arr
     copyBa (Block ba) = primitive $ \s1 -> (# compatCopyByteArrayToAddr# ba os# dst# szBytes# s1, () #)
     copyPtr fptr = unsafePrimFromIO $ withFinalPtr fptr $ \ptr -> copyBytes dst (ptr `plusPtr` os) szBytes
 
+-- | Get a Ptr pointing to the data in the UArray.
+--
+-- Since a UArray is immutable, this Ptr shouldn't be
+-- to use to modify the contents
+--
+-- If the UArray is pinned, then its address is returned as is,
+-- however if it's unpinned, a pinned copy of the UArray is made
+-- before getting the address.
 withPtr :: forall ty prim a . (PrimMonad prim, PrimType ty)
         => UArray ty
         -> (Ptr ty -> prim a)
         -> prim a
-withPtr a f
-    | isPinned a == Pinned =
-        onBackendPrim (\blk  -> BLK.withPtr  blk  $ \ptr -> f (ptr `plusPtr` os))
-                      (\fptr -> withFinalPtr fptr $ \ptr -> f (ptr `plusPtr` os))
-                      a
-    | otherwise = do
-        arr <- do
-            trampoline <- newPinned (length a)
-            unsafeCopyAtRO trampoline 0 a 0 (length a)
-            unsafeFreeze trampoline
-        withPtr arr f
+withPtr a f =
+    onBackendPrim (\blk  -> BLK.withPtr  blk  $ \ptr -> f (ptr `plusPtr` os))
+                  (\fptr -> withFinalPtr fptr $ \ptr -> f (ptr `plusPtr` os))
+                  a
   where
     !sz          = primSizeInBytes (Proxy :: Proxy ty)
     !(Offset os) = offsetOfE sz $ offset a
@@ -329,6 +342,12 @@ recast array
     (CountOf alen) = sizeInBytes (length array)
     missing = alen `mod` bs
 
+-- | Unsafely recast an UArray containing 'a' to an UArray containing 'b'
+--
+-- The offset and size are converted from units of 'a' to units of 'b',
+-- but no check are performed to make sure this is compatible.
+--
+-- use 'recast' if unsure.
 unsafeRecast :: (PrimType a, PrimType b) => UArray a -> UArray b
 unsafeRecast (UArray start len backend) = UArray (primOffsetRecast start) (sizeRecast len) $
     case backend of
@@ -372,14 +391,13 @@ splitAt nbElems arr@(UArray start len backend)
 
 breakElem :: PrimType ty => ty -> UArray ty -> (UArray ty, UArray ty)
 breakElem !ty arr@(UArray start len backend)
--- TODO: return Maybe k
-    | k == end   = (arr, empty)
-    | k == start = (empty, arr)
-    | otherwise  = ( UArray start (offsetAsSize k `sizeSub` offsetAsSize start) backend
-                   , UArray k     (len `sizeSub` (offsetAsSize k `sizeSub` offsetAsSize start)) backend)
+    | k == sentinel = (arr, empty)
+    | k == start    = (empty, arr)
+    | otherwise     = (UArray start (offsetAsSize l1)       backend
+                     , UArray k     (sizeAsOffset len - l1) backend)
   where
-    !end = start `offsetPlusE` len
     !k = onBackendPure' arr $ Alg.findIndexElem ty
+    l1 = k `offsetSub` start
 {-# NOINLINE [3] breakElem #-}
 {-# RULES "breakElem Word8" [4] breakElem = breakElemByte #-}
 {-# SPECIALIZE [3] breakElem :: Word32 -> UArray Word32 -> (UArray Word32, UArray Word32) #-}
@@ -418,11 +436,11 @@ breakLine arr@(UArray start len backend)
     carriageReturn = 0xd
     goBa (Block ba) =
         let k = sysHsMemFindByteBa ba start end lineFeed
-            cr = k > start && PrimBA.primIndex ba (k `offsetSub` 1) == carriageReturn
+            cr = k > start && primBaIndex ba (k `offsetSub` 1) == carriageReturn
          in (if cr then k `offsetSub` 1 else k, k)
     goAddr (Ptr addr) =
         let k = sysHsMemFindByteAddr addr start end lineFeed
-            cr = k > start && PrimAddr.primIndex addr (k `offsetSub` 1) == carriageReturn
+            cr = k > start && primAddrIndex addr (k `offsetSub` 1) == carriageReturn
          in (if cr then k `offsetSub` 1 else k, k)
 
 -- inverse a CountOf that is specified from the end (e.g. take n elements from the end)
@@ -474,35 +492,26 @@ sub (UArray start len backend) startIdx expectedEndIdx
 
 findIndex :: PrimType ty => ty -> UArray ty -> Maybe (Offset ty)
 findIndex ty arr
--- TODO: check for end could be done in algorithm
-    | k == end  = Nothing
-    | otherwise = Just (k `offsetSub` start)
+    | k == sentinel  = Nothing
+    | otherwise      = Just (k `offsetSub` offset arr)
   where
     !k = onBackendPure' arr $ Alg.findIndexElem ty
-    !start = offset arr
-    !end = start `offsetPlusE` length arr
 {-# SPECIALIZE [3] findIndex :: Word8 -> UArray Word8 -> Maybe (Offset Word8) #-}
 
 revFindIndex :: PrimType ty => ty -> UArray ty -> Maybe (Offset ty)
 revFindIndex ty arr
--- TODO: check for end could be done in algorithm
-    | k == end  = Nothing
-    | otherwise = Just (k `offsetSub` start)
+    | k == sentinel = Nothing
+    | otherwise     = Just (k `offsetSub` offset arr)
   where
     !k = onBackendPure' arr $ Alg.revFindIndexElem ty
-    !start = offset arr
-    !end = start `offsetPlusE` length arr
 {-# SPECIALIZE [3] revFindIndex :: Word8 -> UArray Word8 -> Maybe (Offset Word8) #-}
 
 break :: forall ty . PrimType ty => (ty -> Bool) -> UArray ty -> (UArray ty, UArray ty)
 break predicate arr
--- TODO2: check for end could be done in algorithm? but maybe more ops are involved
-    | k == end  = (arr, mempty)
-    | otherwise = splitAt (offsetAsSize (k `offsetSub` start)) arr
+    | k == sentinel = (arr, mempty)
+    | otherwise     = splitAt (k - offset arr) arr
   where
     !k = onBackendPure' arr $ Alg.findIndexPredicate predicate
-    !start = offset arr
-    !end = start `offsetPlusE` length arr
 
 {-
 {-# SPECIALIZE [3] findIndex :: Word8 -> UArray Word8 -> Maybe (Offset Word8) #-}
@@ -535,22 +544,14 @@ break predicate arr
 -- ([1,2,3], [0,0,0])
 breakEnd :: forall ty . PrimType ty => (ty -> Bool) -> UArray ty -> (UArray ty, UArray ty)
 breakEnd predicate arr
--- TODO2: check for end could be done in algorithm? but maybe more ops are involved
-    | k == end  = (arr, mempty)
-    | otherwise = splitAt (offsetAsSize (k+1) `sizeSub` offsetAsSize start) arr
+    | k == sentinel = (arr, mempty)
+    | otherwise     = splitAt ((k+1) - offset arr) arr
   where
     !k = onBackendPure' arr $ Alg.revFindIndexPredicate predicate
-    !start = offset arr
-    !end   = start `offsetPlusE` length arr
 {-# SPECIALIZE [3] breakEnd :: (Word8 -> Bool) -> UArray Word8 -> (UArray Word8, UArray Word8) #-}
 
 elem :: PrimType ty => ty -> UArray ty -> Bool
---elem !ty arr = onBackendPure goBa goAddr arr /= end
--- check for end could be done in algorithm? isNothing?
-elem !ty arr = onBackendPure' arr (Alg.findIndexElem ty) /= end
-  where
-    !start = offset arr
-    !end = start `offsetPlusE` length arr
+elem !ty arr = onBackendPure' arr (Alg.findIndexElem ty) /= sentinel
 {-# SPECIALIZE [2] elem :: Word8 -> UArray Word8 -> Bool #-}
 
 intersperse :: forall ty . PrimType ty => ty -> UArray ty -> UArray ty
